import boto3

from .translator.arn_generator import ArnGenerator, NoRegionFound


class RegionConfiguration:
    """
    There are times when certain services, or certain configurations of a service are not supported in a region. This
    class abstracts all region/partition specific configuration.
    """

    @classmethod
    def is_apigw_edge_configuration_supported(cls) -> bool:
        """
        # API Gateway defaults to EDGE endpoint configuration in all regions in AWS partition. But for other partitions,
        # such as GovCloud, they don't support Edge.

        :return: True, if API Gateway does not support Edge configuration
        """

        return ArnGenerator.get_partition_name() not in [
            "aws-us-gov",
            "aws-iso",
            "aws-iso-b",
            "aws-cn",
        ]

    @classmethod
    def is_service_supported(cls, service, region=None):  # type: ignore[no-untyped-def]
        """
        Not all services are supported in all regions.  This method returns whether a given
        service is supported in a given region.  If no region is specified, the current region
        (as identified by boto3) is used.
        https://aws.amazon.com/about-aws/global-infrastructure/regional-product-services/

        :param service: service code (string used to obtain a boto3 client for the service)
        :param region: region identifier (e.g., us-east-1)
        :return: True, if the service is supported in the region
        """

        # Attempt to re-use an existing session if present.
        session = boto3.Session() if not boto3.DEFAULT_SESSION else boto3.DEFAULT_SESSION

        if not region:
            # get the current region
            region = session.region_name

            # need to handle when region is None so that it won't break
            if region is None:
                if ArnGenerator.BOTO_SESSION_REGION_NAME is not None:
                    region = ArnGenerator.BOTO_SESSION_REGION_NAME
                else:
                    raise NoRegionFound("AWS Region cannot be found")

        # check if the service is available in region
        partition = ArnGenerator.get_partition_name(region)
        available_regions = session.get_available_regions(service, partition_name=partition)
        return region in available_regions



import json
from functools import partial
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, TypeVar, Union

from samtranslator.compat import pydantic
from samtranslator.model.types import PassThrough


# If using PassThrough as-is, pydantic will mark the field as not required:
#  - https://github.com/pydantic/pydantic/issues/990
#  - https://github.com/pydantic/pydantic/issues/1223
#
# That isn't what we want; we want it to specify any type, but still required.
# Using a class gets around it.
class PassThroughProp(pydantic.BaseModel):
    __root__: PassThrough


# Intrinsic resolvable by the SAM transform
T = TypeVar("T")
SamIntrinsicable = Union[Dict[str, Any], T]
SamIntrinsic = Dict[str, Any]

# TODO: Get rid of this in favor of proper types
Unknown = Optional[Any]

DictStrAny = Dict[str, Any]

LenientBaseModel = pydantic.BaseModel

_docdir = Path(__file__).absolute().parent
_DOCS = json.loads((_docdir / "sam-docs.json").read_bytes())


# Connector Permissions
PermissionsType = List[Literal["Read", "Write"]]


def get_prop(stem: str) -> Any:
    return partial(_get_prop, stem)


def passthrough_prop(sam_docs_stem: str, sam_docs_name: str, prop_path: List[str]) -> Any:
    """
    Specifies a pass-through field, where resource_type is the CloudFormation
    resource type, and path is the list of keys to the property.
    """
    path = ["definitions", prop_path[0]]
    for s in prop_path[1:]:
        path.extend(["properties", s])
    docs = _DOCS["properties"][sam_docs_stem][sam_docs_name]
    return pydantic.Field(
        title=sam_docs_name,
        # We add a custom value to the schema containing the path to the pass-through
        # documentation; the dict containing the value is replaced in the final schema
        __samPassThrough={
            # To know at schema build-time where to find the property schema
            "schemaPath": path,
            # Use SAM docs at the top-level pass-through; it can include useful SAM-specific information
            "markdownDescriptionOverride": docs,
        },
    )


def _get_prop(stem: str, name: str) -> Any:
    docs = _DOCS["properties"][stem][name]
    return pydantic.Field(
        title=name,
        # https://code.visualstudio.com/docs/languages/json#_use-rich-formatting-in-hovers
        markdownDescription=docs,
    )


# By default strict: https://pydantic-docs.helpmanual.io/usage/model_config/#change-behaviour-globally
class BaseModel(LenientBaseModel):
    class Config:
        extra = pydantic.Extra.forbid

    def __getattribute__(self, __name: str) -> Any:
        """Overloading get attribute operation to allow access PassThroughProp without using __root__"""
        attr_value = super().__getattribute__(__name)
        if isinstance(attr_value, PassThroughProp):
            # See docstring of PassThroughProp
            return attr_value.__root__
        return attr_value


# https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-ref.html
class Ref(BaseModel):
    Ref: str


class ResourceAttributes(BaseModel):
    DependsOn: Optional[PassThroughProp]
    DeletionPolicy: Optional[PassThroughProp]
    Metadata: Optional[PassThroughProp]
    UpdateReplacePolicy: Optional[PassThroughProp]
    Condition: Optional[PassThroughProp]
    IgnoreGlobals: Optional[Union[str, List[str]]]



from functools import lru_cache
from typing import Optional

import boto3


class NoRegionFound(Exception):
    pass


@lru_cache(maxsize=1)  # Only need to cache one as once deployed, it is not gonna deal with another region.
def _get_region_from_session() -> str:
    return boto3.session.Session().region_name


@lru_cache(maxsize=1)  # Only need to cache one as once deployed, it is not gonna deal with another region.
def _region_to_partition(region: str) -> str:
    # setting default partition to aws, this will be overwritten by checking the region below
    region_string = region.lower()
    if region_string.startswith("cn-"):
        return "aws-cn"
    if region_string.startswith("us-iso-"):
        return "aws-iso"
    if region_string.startswith("us-isob"):
        return "aws-iso-b"
    if region_string.startswith("us-gov"):
        return "aws-us-gov"

    return "aws"


class ArnGenerator:
    BOTO_SESSION_REGION_NAME: Optional[str] = None

    @classmethod
    def generate_arn(
        cls,
        partition: str,
        service: str,
        resource: str,
        include_account_id: bool = True,
        region: Optional[str] = None,
    ) -> str:
        """Generate AWS ARN.

        Parameters
        ----------
        partition
            AWS partition, ie "aws" or "aws-cn"
        service
            AWS service name
        resource
            Resource name, it must include service specific prefixes is "table/" for DynamoDB table
        include_account_id, optional
            include account ID in the ARN or not, by default True
        region, optional
            resource region, by default None.
            To omit region in ARN (ie for a S3 bucket) pass "" (empty string).
            Don't set it to any other default value because None can be passed by a caller and
            must be handled in the function itself.

        Returns
        -------
            Generated ARN

        Raises
        ------
        RuntimeError
            if service or resource are not provided
        """
        if not service or not resource:
            raise RuntimeError("Could not construct ARN for resource.")

        if region is None:
            region = "${AWS::Region}"

        arn = "arn:{0}:{1}:{region}:"

        if include_account_id:
            arn += "${{AWS::AccountId}}:"

        arn += "{2}"

        return arn.format(partition, service, resource, region=region)

    @classmethod
    def generate_aws_managed_policy_arn(cls, policy_name: str) -> str:
        """
        Method to create an ARN of AWS Owned Managed Policy. This uses the right partition name to construct
        the ARN

        :param policy_name: Name of the policy
        :return: ARN Of the managed policy
        """
        return f"arn:{ArnGenerator.get_partition_name()}:iam::aws:policy/{policy_name}"

    @classmethod
    def get_partition_name(cls, region: Optional[str] = None) -> str:
        """
        Gets the name of the partition given the region name. If region name is not provided, this method will
        use Boto3 to get name of the region where this code is running.

        This implementation is borrowed from AWS CLI
        https://github.com/aws/aws-cli/blob/1.11.139/awscli/customizations/emr/createdefaultroles.py#L59

        :param region: Optional name of the region
        :return: Partition name
        """

        if region is None:
            # Use Boto3 to get the region where code is running. This uses Boto's regular region resolution
            # mechanism, starting from AWS_DEFAULT_REGION environment variable.

            region = (
                _get_region_from_session()
                if ArnGenerator.BOTO_SESSION_REGION_NAME is None
                else ArnGenerator.BOTO_SESSION_REGION_NAME
            )

        # If region is still None, then we could not find the region. This will only happen
        # in the local context. When this is deployed, we will be able to find the region like
        # we did before.
        if region is None:
            raise NoRegionFound("AWS Region cannot be found")

        return _region_to_partition(region)

    @classmethod
    def generate_dynamodb_table_arn(cls, partition: str, region: str, table_name: str) -> str:
        """Generate DynamoDB table ARN.

        Parameters
        ----------
        partition
            _description_
        region
            DynamoDB table region
        table_name
            DynamoDB table name

        Returns
        -------
            DynamoDB table ARN.
        """
        return ArnGenerator.generate_arn(
            partition=partition, service="dynamodb", resource=f"table/{table_name}", region=region
        )



import re
from typing import Any


def is_dynamic_reference(_input: Any) -> bool:
    """
    Checks if the given input is a dynamic reference. Dynamic references follow the pattern '{{resolve:service-name:reference-key}}'

    This method does not validate if the dynamic reference is valid or not, only if it follows the valid pattern: {{resolve:service-name:reference-key}}

    :param _input: Input value to check if it is a dynamic reference
    :return: True, if yes
    """
    pattern = re.compile("^{{resolve:([a-z-]+):(.+)}}$")
    if _input is not None and isinstance(_input, str) and pattern.match(_input):
        return True
    return False



import logging
from typing import Any, Dict

from samtranslator.model.exceptions import (
    InvalidDocumentException,
    InvalidResourceAttributeTypeException,
    InvalidTemplateException,
)
from samtranslator.plugins import LifeCycleEvents
from samtranslator.plugins.sam_plugins import SamPlugins
from samtranslator.public.sdk.template import SamTemplate
from samtranslator.validator.value_validator import sam_expect

LOG = logging.getLogger(__name__)


class Parser:
    def __init__(self) -> None:
        pass

    def parse(self, sam_template: Dict[str, Any], parameter_values: Dict[str, Any], sam_plugins: SamPlugins) -> None:
        self._validate(sam_template, parameter_values)  # type: ignore[no-untyped-call]
        sam_plugins.act(LifeCycleEvents.before_transform_template, sam_template)

    @staticmethod
    def validate_datatypes(sam_template):  # type: ignore[no-untyped-def]
        """Validates the datatype within the template"""
        if (
            "Resources" not in sam_template
            or not isinstance(sam_template["Resources"], dict)
            or not sam_template["Resources"]
        ):
            raise InvalidDocumentException([InvalidTemplateException("'Resources' section is required")])

        if not all(isinstance(sam_resource, dict) for sam_resource in sam_template["Resources"].values()):
            raise InvalidDocumentException(
                [
                    InvalidTemplateException(
                        "All 'Resources' must be Objects. If you're using YAML, this may be an indentation issue."
                    )
                ]
            )

        sam_template_instance = SamTemplate(sam_template)

        for resource_logical_id, sam_resource in sam_template_instance.iterate():
            # NOTE: Properties isn't required for SimpleTable, so we can't check
            # `not isinstance(sam_resources.get("Properties"), dict)` as this would be a breaking change.
            # sam_resource.properties defaults to {} in SamTemplate init
            try:
                sam_expect(
                    sam_resource.properties, resource_logical_id, "Properties", is_resource_attribute=True
                ).to_be_a_map()
            except InvalidResourceAttributeTypeException as e:
                raise InvalidDocumentException([e]) from e

    # private methods
    def _validate(self, sam_template, parameter_values):  # type: ignore[no-untyped-def]
        """Validates the template and parameter values and raises exceptions if there's an issue

        :param dict sam_template: SAM template
        :param dict parameter_values: Dictionary of parameter values provided by the user
        """
        if parameter_values is None:
            raise ValueError("`parameter_values` argument is required")

        Parser.validate_datatypes(sam_template)  # type: ignore[no-untyped-call]



# Help resolve intrinsic functions
from typing import Any, Callable, Dict, List, Optional, Union, cast

from samtranslator.intrinsics.actions import Action, GetAttAction, RefAction, SubAction
from samtranslator.intrinsics.resource_refs import SupportedResourceReferences
from samtranslator.model.exceptions import InvalidDocumentException, InvalidTemplateException

# All intrinsics are supported by default
DEFAULT_SUPPORTED_INTRINSICS = {action.intrinsic_name: action() for action in [RefAction, SubAction, GetAttAction]}


class IntrinsicsResolver:
    def __init__(self, parameters: Dict[str, Any], supported_intrinsics: Optional[Dict[str, Any]] = None) -> None:
        """
        Instantiate the resolver
        :param dict parameters: Map of parameter names to their values
        :param dict supported_intrinsics: Dictionary of intrinsic functions this class supports along with the
            Action class that can process this intrinsic
        :raises TypeError: If parameters or the supported_intrinsics arguments are invalid
        """

        if supported_intrinsics is None:
            supported_intrinsics = DEFAULT_SUPPORTED_INTRINSICS
        if parameters is None or not isinstance(parameters, dict):
            raise InvalidDocumentException(
                [InvalidTemplateException("'Mappings' or 'Parameters' is either null or not a valid dictionary.")]
            )

        if not isinstance(supported_intrinsics, dict) or not all(
            isinstance(value, Action) for value in supported_intrinsics.values()
        ):
            raise TypeError("supported_intrinsics argument must be intrinsic names to corresponding Action classes")

        self.supported_intrinsics = supported_intrinsics
        self.parameters = parameters

    def resolve_parameter_refs(self, _input: Any) -> Any:
        """
        Resolves references to parameters within the given dictionary recursively. Other intrinsic functions such as
        !GetAtt, !Sub or !Ref to non-parameters will be left untouched.

        Result is a dictionary where parameter values are inlined. Don't pass this dictionary directly into
        transform's output because it changes the template structure by inlining parameter values.

        :param _input: Any primitive type (dict, array, string etc) whose values might contain intrinsic functions
        :return: A copy of a dictionary with parameter references replaced by actual value.
        """
        return self._traverse(_input, self.parameters, self._try_resolve_parameter_refs)

    def resolve_sam_resource_refs(
        self, _input: Dict[str, Any], supported_resource_refs: SupportedResourceReferences
    ) -> Dict[str, Any]:
        """
        Customers can provide a reference to a "derived" SAM resource such as Alias of a Function or Stage of an API
        resource. This method recursively walks the tree, converting all derived references to the real resource name,
        if it is present.

        Example:
            {"Ref": "MyFunction.Alias"} -> {"Ref": "MyFunctionAliasLive"}

        This method does not attempt to validate a reference. If it is invalid or non-resolvable, it skips the
        occurrence and continues with the rest. It is recommended that you have an external process that detects and
        surfaces invalid references.

        For first call, it is recommended that `template` is the entire CFN template in order to handle
        references in Mapping or Output sections.

        :param dict input: CFN template that needs resolution. This method will modify the input
            directly resolving references. In subsequent recursions, this will be a fragment of the CFN template.
        :param SupportedResourceReferences supported_resource_refs: Object that contains information about the resource
            references supported in this SAM template, along with the value they should resolve to.
        :return list errors: List of dictionary containing information about invalid reference. Empty list otherwise
        """
        # The _traverse() return type is the same as the input. Here the input is Dict[str, Any]
        return cast(
            Dict[str, Any], self._traverse(_input, supported_resource_refs, self._try_resolve_sam_resource_refs)
        )

    def resolve_sam_resource_id_refs(self, _input: Dict[str, Any], supported_resource_id_refs: Dict[str, str]) -> Any:
        """
        Some SAM resources have their logical ids mutated from the original id that the customer writes in the
        template. This method recursively walks the tree and updates these logical ids from the old value
        to the new value that is generated by SAM.

        Example:
            {"Ref": "MyLayer"} -> {"Ref": "MyLayerABC123"}

        This method does not attempt to validate a reference. If it is invalid or non-resolvable, it skips the
        occurrence and continues with the rest. It is recommended that you have an external process that detects and
        surfaces invalid references.

        For first call, it is recommended that `template` is the entire CFN template in order to handle
        references in Mapping or Output sections.

        :param dict input: CFN template that needs resolution. This method will modify the input
            directly resolving references. In subsequent recursions, this will be a fragment of the CFN template.
        :param dict supported_resource_id_refs: Dictionary that maps old logical ids to new ones.
        :return list errors: List of dictionary containing information about invalid reference. Empty list otherwise
        """
        return self._traverse(_input, supported_resource_id_refs, self._try_resolve_sam_resource_id_refs)

    def _traverse(
        self,
        input_value: Any,
        resolution_data: Union[Dict[str, Any], SupportedResourceReferences],
        resolver_method: Callable[[Dict[str, Any], Any], Any],
    ) -> Any:
        """
        Driver method that performs the actual traversal of input and calls the appropriate `resolver_method` when
        to perform the resolution.

        :param input_value: Any primitive type  (dict, array, string etc) whose value might contain an intrinsic function
        :param resolution_data: Data that will help with resolution. For example, when resolving parameter references,
            this object will contain a dictionary of parameter names and their values.
        :param resolver_method: Method that will be called to actually resolve an intrinsic function. This method
            is called with the parameters `(input, resolution_data)`.
        :return: Modified `input` with intrinsics resolved

        TODO: type this and make _traverse generic.
        """

        # There is data to help with resolution. Skip the traversal altogether
        if len(resolution_data) == 0:
            return input_value

        #
        # Traversal Algorithm:
        #
        # Imagine the input dictionary/list as a tree. We are doing a Pre-Order tree traversal here where we first
        # process the root node before going to its children. Dict and Lists are the only two iterable nodes.
        # Everything else is a leaf node.
        #
        # We do a Pre-Order traversal to handle the case where `input` contains intrinsic function as its only child
        # ie. input = {"Ref": "foo}.
        #
        # We will try to resolve the intrinsics if we can, otherwise return the original input. In some cases, resolving
        # an intrinsic will result in a terminal state ie. {"Ref": "foo"} could resolve to a string "bar". In other
        # cases, resolving intrinsics is only partial and we might need to continue traversing the tree (ex: Fn::Sub)
        # to handle nested intrinsics. All of these cases lend well towards a Pre-Order traversal where we try and
        # process the intrinsic, which results in a modified sub-tree to traverse.
        #
        input_value = resolver_method(input_value, resolution_data)
        if isinstance(input_value, dict):
            return self._traverse_dict(input_value, resolution_data, resolver_method)
        if isinstance(input_value, list):
            return self._traverse_list(input_value, resolution_data, resolver_method)
        # We can iterate only over dict or list types. Primitive types are terminals

        return input_value

    def _traverse_dict(
        self,
        input_dict: Dict[str, Any],
        resolution_data: Union[Dict[str, Any], SupportedResourceReferences],
        resolver_method: Callable[[Dict[str, Any], Any], Any],
    ) -> Any:
        """
        Traverse a dictionary to resolve intrinsic functions on every value

        :param input_dict: Input dictionary to traverse
        :param resolution_data: Data that the `resolver_method` needs to operate
        :param resolver_method: Method that can actually resolve an intrinsic function, if it detects one
        :return: Modified dictionary with values resolved
        """
        for key, value in input_dict.items():
            input_dict[key] = self._traverse(value, resolution_data, resolver_method)

        return input_dict

    def _traverse_list(
        self,
        input_list: List[Any],
        resolution_data: Union[Dict[str, Any], SupportedResourceReferences],
        resolver_method: Callable[[Dict[str, Any], Any], Any],
    ) -> Any:
        """
        Traverse a list to resolve intrinsic functions on every element

        :param input_list: List of input
        :param resolution_data: Data that the `resolver_method` needs to operate
        :param resolver_method: Method that can actually resolve an intrinsic function, if it detects one
        :return: Modified list with intrinsic functions resolved
        """
        for index, value in enumerate(input_list):
            input_list[index] = self._traverse(value, resolution_data, resolver_method)

        return input_list

    def _try_resolve_parameter_refs(self, _input: Dict[str, Any], parameters: Dict[str, Any]) -> Any:
        """
        Try to resolve parameter references on the given input object. The object could be of any type.
        If the input is not in the format used by intrinsics (ie. dictionary with one key), input is returned
        unmodified. If the single key in dictionary is one of the supported intrinsic function types,
        go ahead and try to resolve it.

        :param _input: Input object to resolve
        :param parameters: Parameter values used to for ref substitution
        :return:
        """
        if not self._is_intrinsic_dict(_input):
            return _input

        function_type = next(iter(_input.keys()))
        return self.supported_intrinsics[function_type].resolve_parameter_refs(_input, parameters)

    def _try_resolve_sam_resource_refs(
        self, _input: Dict[str, Any], supported_resource_refs: SupportedResourceReferences
    ) -> Any:
        """
        Try to resolve SAM resource references on the given template. If the given object looks like one of the
        supported intrinsics, it calls the appropriate resolution on it. If not, this method returns the original input
        unmodified.

        :param dict _input: Dictionary that may represent an intrinsic function
        :param SupportedResourceReferences supported_resource_refs: Object containing information about available
            resource references and the values they resolve to.
        :return: Modified input dictionary with references resolved
        """
        if not self._is_intrinsic_dict(_input):
            return _input

        function_type = next(iter(_input.keys()))
        return self.supported_intrinsics[function_type].resolve_resource_refs(_input, supported_resource_refs)

    def _try_resolve_sam_resource_id_refs(
        self, _input: Dict[str, Any], supported_resource_id_refs: Dict[str, str]
    ) -> Any:
        """
        Try to resolve SAM resource id references on the given template. If the given object looks like one of the
        supported intrinsics, it calls the appropriate resolution on it. If not, this method returns the original input
        unmodified.

        :param dict _input: Dictionary that may represent an intrinsic function
        :param dict supported_resource_id_refs: Dictionary that maps old logical ids to new ones.
        :return: Modified input dictionary with id references resolved
        """
        if not self._is_intrinsic_dict(_input):
            return _input

        function_type = next(iter(_input.keys()))
        return self.supported_intrinsics[function_type].resolve_resource_id_refs(_input, supported_resource_id_refs)

    def _is_intrinsic_dict(self, _input: Dict[str, Any]) -> bool:
        """
        Can the _input represent an intrinsic function in it?

        :param _input: Object to be checked
        :return: True, if the _input contains a supported intrinsic function.  False otherwise
        """
        # All intrinsic functions are dictionaries with just one key
        return isinstance(_input, dict) and len(_input) == 1 and next(iter(_input.keys())) in self.supported_intrinsics
